@@ -18,6 +18,36 @@ from .introspection import (
     Cardinality,
 )
 
+
+class DBVendor(str, Enum):
+    edgedb = "edgedb"
+
+
+class TLSSecurity(str, Enum):
+    insecure = "insecure"
+
+
+class PropertyConfig(BaseModel):
+    module_path: str
+    module_name: str
+
+
+class NodeConfig(BaseModel):
+    module_path: str = None
+    appendix_properties: T.List[str] = []
+    basemodel_properties: T.Dict[str, PropertyConfig] = {}
+
+
+class DBConfig(BaseModel):
+    vendor: DBVendor
+    host: str
+    password: str
+    port: int = 5656
+    tls_security: TLSSecurity
+    hydrate: bool = False
+    nodes: T.Dict[str, NodeConfig] = dict()
+
+
 PATH_TO_MODULE = "edgedb_orm"
 DEFAULT_INDENT = "    "
 CONFIG_NAME = "GraphORM"
@@ -41,8 +71,8 @@ def imports() -> str:
         "from uuid import UUID",
         "from decimal import Decimal",
         "from edgedb import RelativeDuration, AsyncIOClient, create_async_client",
-        "from pydantic import BaseModel, Field, PrivateAttr",
-        f"from {PATH_TO_MODULE} import Node, Resolver, NodeException, ResolverException, UpdateOperation, Batch, Unset, ComputedPropertyException, is_unset",
+        "from pydantic import BaseModel, Field, PrivateAttr, validator",
+        f"from {PATH_TO_MODULE} import Node, Resolver, NodeException, ResolverException, UpdateOperation, Batch, Unset, ComputedPropertyException, AppendixPropertyException, is_unset, from_str, enum_from_str",
         'EdgeResolverType = T.TypeVar("EdgeResolverType", bound=Resolver)',
     ]
     return "\n".join(lines)
@@ -184,7 +214,10 @@ def stringify_dict(
 
 
 def stringify_set(s: T.Set[str]) -> str:
-    strs: T.List[str] = [f'"{i}"' for i in s]
+    if not s:
+        return "set()"
+    s_sorted = sorted(list(s))
+    strs: T.List[str] = [f'"{i}"' for i in s_sorted]
     return "{" + ",".join(strs) + "}"
 
 
@@ -195,8 +228,46 @@ def edgedb_conversion_type_from_prop(prop: Property) -> str:
     return s
 
 
+def build_validator_module_imports(db_config: DBConfig) -> str:
+    if not db_config.nodes:
+        return ""
+    node_configs = db_config.nodes.values()
+    from_str_import_strs: T.List[str] = []
+    for node_config in node_configs:
+        for prop_name, prop_config in node_config.basemodel_properties.items():
+            from_str_import_strs.append(
+                f"from {prop_config.module_path} import {prop_config.module_name} as {prop_config.module_name}Model"
+            )
+    return "\n".join(sorted(list(set(from_str_import_strs))))
+
+
+def build_from_str_validator_str(node_config: T.Optional[NodeConfig]) -> str:
+    if not node_config:
+        return ""
+    field_names = node_config.basemodel_properties.keys()
+    if not field_names:
+        return ""
+    field_name_strs: T.List[str] = []
+    for field_name in field_names:
+        key = field_name
+        if field_name in node_config.appendix_properties:
+            key += "_"
+        field_name_strs.append(f'"{key}"')
+    return f"""
+_from_str = validator({", ".join(field_name_strs)}, pre=True, allow_reuse=True)(from_str)
+    """
+
+
+def remove_falsies(lst: list) -> list:
+    return [i for i in lst if i]
+
+
 def build_node_and_resolver(
-    object_type: ObjectType, edge_resolver_map_strs: T.List[str], hydrate: bool
+    object_type: ObjectType,
+    node_config: T.Optional[NodeConfig],
+    edge_resolver_map_strs: T.List[str],
+    hydrate: bool,
+    dehydrate: bool,
 ) -> str:
     # need to sort props and links by required, exclusive, no default, rest
     object_type.properties.sort(
@@ -220,6 +291,8 @@ def build_node_and_resolver(
     computed_properties: T.Set[str] = set()
     computed_property_getter_strs: T.List[str] = []
 
+    appendix_properties: T.Set[str] = set()
+
     for prop in object_type.properties:
         conversion_type = edgedb_conversion_type_from_prop(prop)
         node_edgedb_conversion_map[prop.name] = {
@@ -229,6 +302,11 @@ def build_node_and_resolver(
         }
         if prop.is_computed:
             computed_properties.add(prop.name)
+        is_appendix = False
+        if node_config:
+            if prop.name in node_config.appendix_properties:
+                is_appendix = True
+                appendix_properties.add(prop.name)
         if not prop.readonly and not prop.is_computed:
             updatable_fields.add(prop.name)
         if prop.is_exclusive:
@@ -237,26 +315,45 @@ def build_node_and_resolver(
         allow_mutation_str = (
             f"allow_mutation={not prop.readonly and not prop.is_computed}"
         )
-        if not prop.is_computed:
+        if node_config and prop.name in node_config.basemodel_properties:
+            module_name = node_config.basemodel_properties[prop.name].module_name
+            type_str = prop.type_str_basemodel(module_name)
+        else:
+            type_str = prop.type_str
+        if not (prop.is_computed or is_appendix):
             property_strs.append(
-                f"{prop.name}: {prop.type_str} = Field({default_value_str}, {allow_mutation_str})"
+                f"{prop.name}: {type_str} = Field({default_value_str}, {allow_mutation_str})"
             )
         else:
-            property_strs.append(
-                f"_{prop.name}: T.Union[Unset, {prop.type_str}] = PrivateAttr(default_factory=Unset)"
-            )
+            if prop.is_computed:
+                exception_name = "ComputedPropertyException"
+                property_name = f"_{prop.name}"
+                property_str = f"{property_name}: T.Union[{type_str}, Unset] = PrivateAttr(default_factory=Unset)"
+            else:
+                exception_name = "AppendixPropertyException"
+                property_name = f"{prop.name}_"
+                property_str = f'{property_name}: T.Union[{type_str}, Unset] = Field(default_factory=Unset, alias="{prop.name}")'
+            property_strs.append(property_str)
             computed_property_getter_strs.append(
                 f"""
 @property
-def {prop.name}(self) -> {prop.type_str}:
-    if is_unset(self._{prop.name}):
+def {prop.name}(self) -> {type_str}:
+    if is_unset(self.{property_name}):
         if "{prop.name}" in self.extra:
-            self._{prop.name} = self.extra["{prop.name}"]
+            self.{property_name} = self.extra["{prop.name}"]
         else:
-            raise ComputedPropertyException("{prop.name} is unset")
-    return self._{prop.name}
+            raise {exception_name}("{prop.name} is unset")
+    return self.{property_name}
                 """
             )
+            if not prop.is_computed:
+                computed_property_getter_strs.append(
+                    f"""
+@{prop.name}.setter
+def {prop.name}(self, {prop.name}: {type_str}) -> None:
+    self.{property_name} = {prop.name}
+                    """
+                )
         # for insert type
         if prop.name != "id":
             if not prop.is_computed and not prop.not_insertable:
@@ -265,7 +362,7 @@ def {prop.name}(self) -> {prop.type_str}:
                     "cardinality": prop.cardinality.value,
                     "readonly": prop.readonly,
                 }
-                insert_type_str = prop.type_str
+                insert_type_str = type_str
                 # if required but has default, add optional back
                 if prop.required and prop.default:
                     insert_type_str = f"T.Optional[{insert_type_str}]"
@@ -328,36 +425,51 @@ def {prop.name}(self) -> {prop.type_str}:
 
     # node
     node_properties_str = "\n".join(property_strs)
+    from_str_validator_str = build_from_str_validator_str(node_config=node_config)
     if hydrate:
         node_properties_str = ""
+        from_str_validator_str = ""
+
     computed_property_getter_str = "\n".join(computed_property_getter_strs)
     node_conversion_map_str = f"_edgedb_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(node_edgedb_conversion_map)}"
     insert_link_conversion_map_str = f"_link_conversion_map: T.ClassVar[T.Dict[str, str]] = {stringify_dict(link_conversion_map)}"
     computed_properties_str = f"_computed_properties: T.ClassVar[T.Set[str]] = {stringify_set(computed_properties)}"
+    appendix_properties_str = f"_appendix_properties: T.ClassVar[T.Set[str]] = {stringify_set(appendix_properties)}"
+    basemodel_properties = (
+        [] if not node_config else node_config.basemodel_properties.keys()
+    )
+    basemodel_properties_str = f"_basemodel_properties: T.ClassVar[T.Set[str]] = {stringify_set(set(basemodel_properties))}"
     node_link_functions_str = "\n".join(link_function_strs)
     update_function_str = build_update_function_str(
         node_resolver_name=node_resolver_name, links=object_type.links
     )
+    if dehydrate:
+        node_link_functions_str = ""
+        update_function_str = ""
     node_inner_strs = [
         node_properties_str,
         "\n",
+        from_str_validator_str,
+        "\n",
         computed_property_getter_str,
         node_conversion_map_str,
         insert_link_conversion_map_str,
         computed_properties_str,
+        appendix_properties_str,
+        basemodel_properties_str,
         node_link_functions_str,
         update_function_str,
         orm_config_str,
     ]
-    node_inner_str = "\n".join(node_inner_strs)
+    node_inner_str = "\n".join(remove_falsies(node_inner_strs))
+    node_outer_strs = []
+    node_outer_str = "\n".join(remove_falsies(node_outer_strs))
     inherits = (
         f"Node[{insert_model_name}]"
         if not hydrate
         else f"{object_type.node_name}Hydrated"
     )
-    node_s = (
-        f"class {object_type.node_name}({inherits}):\n{indent_lines(node_inner_str)}"
-    )
+    node_s = f"{node_outer_str}\nclass {object_type.node_name}({inherits}):\n{indent_lines(node_inner_str)}"
 
     # resolver
     resolver_properties_str = f"_node = {object_type.node_name}"
@@ -393,7 +505,10 @@ def {prop.name}(self) -> {prop.type_str}:
 
 
 async def build_nodes_and_resolvers(
-    client: edgedb.AsyncIOClient, nodes_to_hydrate: T.Set[str]
+    client: edgedb.AsyncIOClient,
+    db_config: DBConfig,
+    nodes_to_hydrate: T.Set[str],
+    dehydrate: bool,
 ) -> str:
     object_types = await introspect_objects(client)
     node_strs: T.List[str] = []
@@ -402,38 +517,21 @@ async def build_nodes_and_resolvers(
         node_strs.append(
             build_node_and_resolver(
                 object_type,
+                node_config=db_config.nodes.get(object_type.node_name),
                 edge_resolver_map_strs=edge_resolver_map_strs,
-                hydrate=object_type.node_name in nodes_to_hydrate,
+                hydrate=object_type.node_name in nodes_to_hydrate and not dehydrate,
+                dehydrate=dehydrate,
             )
         )
-    update_forward_refs_str = "\n".join(
+    update_forward_refs_inserts_str = "\n".join(
         [f"{o.node_name}Insert.update_forward_refs()" for o in object_types]
     )
+    update_forward_refs_nodes_str = "\n".join(
+        [f"{o.node_name}.update_forward_refs()" for o in object_types]
+    )
     nodes_str = "\n".join(node_strs)
     edge_resolver_map_str = "\n".join(edge_resolver_map_strs)
-    return f"{nodes_str}\n\n{update_forward_refs_str}\n\n{edge_resolver_map_str}"
-
-
-class DBVendor(str, Enum):
-    edgedb = "edgedb"
-
-
-class TLSSecurity(str, Enum):
-    insecure = "insecure"
-
-
-class NodeConfig(BaseModel):
-    module_path: str
-
-
-class DBConfig(BaseModel):
-    vendor: DBVendor
-    host: str
-    password: str
-    port: int = 5656
-    tls_security: TLSSecurity
-    hydrate: bool = False
-    nodes: T.Dict[str, NodeConfig] = dict()
+    return f"{nodes_str}\n\n{update_forward_refs_inserts_str}\n\n{update_forward_refs_nodes_str}\n\n{edge_resolver_map_str}"
 
 
 def build_client(db_config: DBConfig) -> str:
@@ -481,7 +579,10 @@ def get_nodes_to_hydrate(db_config: DBConfig) -> T.Set[str]:
 
 
 async def build_from_config(
-    db_config_d: dict, include_strawberry: bool, hydrate: bool = False
+    db_config_d: dict,
+    include_strawberry: bool,
+    hydrate: bool = False,
+    dehydrate: bool = False,
 ) -> str:
     db_config = DBConfig.parse_obj(db_config_d)
     client = edgedb.create_async_client(
@@ -492,15 +593,25 @@ async def build_from_config(
     )
     imports_str = imports()
     client_str = build_client(db_config)
+    validator_module_imports = build_validator_module_imports(db_config)
     hydrate_imports = "" if not hydrate else build_hydrate_imports(db_config)
     enums_str = await build_enums(client, include_strawberry=include_strawberry)
     nodes_and_resolvers_str = await build_nodes_and_resolvers(
         client,
-        nodes_to_hydrate=set() if not hydrate else get_nodes_to_hydrate(db_config),
+        db_config=db_config,
+        nodes_to_hydrate=get_nodes_to_hydrate(db_config),
+        dehydrate=dehydrate,
     )
 
     s = "\n".join(
-        [imports_str, client_str, hydrate_imports, enums_str, nodes_and_resolvers_str]
+        [
+            imports_str,
+            client_str,
+            validator_module_imports,
+            hydrate_imports,
+            enums_str,
+            nodes_and_resolvers_str,
+        ]
     )
     s = format_str(s, mode=FileMode())
     return s
@@ -516,6 +627,7 @@ async def generate(
         s = await build_from_config(
             db_config_d=db_config_d,
             include_strawberry=include_strawberry,
+            dehydrate=hydrate,
         )
         open(output_path / f"{db_name}.py", "w").write(s)
         if hydrate:
