@@ -140,15 +140,37 @@ class Node(BaseModel, T.Generic[InsertType]):
     _edgedb_conversion_map: T.ClassVar[T.Dict[str, T.Dict[str, str]]]
     _link_conversion_map: T.ClassVar[T.Dict[str, T.Dict[str, str]]]
     _computed_properties: T.ClassVar[T.Set[str]]
+    _appendix_properties: T.ClassVar[T.Set[str]]
+    _basemodel_properties: T.ClassVar[T.Set[str]]
 
     _extra: dict = PrivateAttr(default=dict())
 
+    __updated_fields__: T.Set[str] = PrivateAttr(default=set())
+
+    def __setattr__(self, key, value):
+        self.__updated_fields__.add(key)
+        if key in self._appendix_properties:
+            # TODO is this the desired effect?
+            #  Probably, if you set an appendix field, add it to fields returned in future
+            self._used_resolver.include_fields(key)
+            k = f"{key}_"
+            if k in self.__fields__.keys():
+                key = k
+        return super().__setattr__(key, value)
+
+    def updated_fields(self, *updated_fields: str) -> None:
+        self.__updated_fields__.update(updated_fields)
+
+    def _clear_updated_fields(self) -> None:
+        self.__updated_fields__ = set()
+
     @property
     def extra(self) -> dict:
         return self._extra
 
     class Config:
         validate_assignment = True
+        underscore_attrs_are_private = False
 
     class GraphORM:
         model_name: str = None
@@ -467,8 +489,31 @@ for item in json_array_unpack(raw_data) union ({insert_str}) {given_resolver.to_
             new_node._deleted = True
             models_by_id[new_node.id].hydrate(new_node=new_node)
 
-    def get_update_d(self) -> dict:
-        current_d = self.dict()
+    def build_update_d(self) -> dict:
+        fields_to_update = self.__updated_fields__ & self.GraphORM.updatable_fields
+        update_d = {}
+        for field in fields_to_update:
+            val = getattr(self, field)
+            if val is None:
+                val = set()
+            elif isinstance(val, BaseModel):
+                val = val.json()
+            update_d[field] = val
+        return update_d
+
+    def get_update_d_old(self) -> dict:
+        """DEP. Keep incase new one ^'build_update_d' is not good"""
+        # for appendix properties
+        # if you want to edit an appendix field, you must request it
+        current_d = {**self.extra, **self.dict()}
+        for field, val in current_d.items():
+            print(f"{field=}, {val=}")
+            if (
+                field in self._basemodel_properties
+                and field in self.GraphORM.updatable_fields
+                and isinstance(getattr(self, field, None), BaseModel)
+            ):
+                current_d[field] = getattr(self, field).json()
         original_d = self._original_dict
         unset_key = "_UNSET_12398120"
         update_d = {}
@@ -478,8 +523,19 @@ for item in json_array_unpack(raw_data) union ({insert_str}) {given_resolver.to_
             if current_val != original_val:
                 if current_val != unset_key:
                     # if current val is not unset -> always set
+                    print(f"{current_val=}, {original_val=}")
                     if current_val is not None:
-                        update_d[field_name] = current_val
+                        if (
+                            isinstance(current_val, datetime)
+                            and current_val.isoformat() == original_val
+                        ):
+                            continue
+                        else:
+                            if isinstance(current_val, datetime):
+                                print(
+                                    f"datetimes diff for update: {current_val.isoformat()=}, {original_val=}"
+                                )
+                            update_d[field_name] = current_val
                     else:
                         update_d[field_name] = set()
                 else:
@@ -498,7 +554,7 @@ for item in json_array_unpack(raw_data) union ({insert_str}) {given_resolver.to_
         """set_links_d -> {people_i_follow += (select Person filter .first_name = <str>first_name)}"""
         if not given_resolver:
             given_resolver = self._used_resolver
-        update_d = self.get_update_d()
+        update_d = self.build_update_d()
         if not update_d:
             message = f"No update registered for {self.id=}."
             if error_if_no_update:
@@ -561,8 +617,12 @@ for item in json_array_unpack(raw_data) union ({insert_str}) {given_resolver.to_
         exclusive_fields.add("id")
         return exclusive_fields
 
-    def hydrate(self, new_node: ThisNodeType) -> None:
+    def hydrate(
+        self, new_node: ThisNodeType, clear_updated_fields: bool = True
+    ) -> None:
         """Turns this node into the new node"""
+        if clear_updated_fields:
+            self._clear_updated_fields()
         for field_name in new_node.__fields__.keys():
             new_field = getattr(new_node, field_name)
             old_field = getattr(self, field_name)
