@@ -65,7 +65,10 @@ class Resolver(BaseModel, T.Generic[NodeType]):
     def get_fields_to_return(self) -> T.Set[str]:
         """Include nested fields with DOT notation"""
         if self._fields_to_return is None:
-            return {*self.get_aliases(), *self._fields_to_include}
+            return {
+                *(self.get_aliases() - self._node._appendix_properties),
+                *self._fields_to_include,
+            }
         return self._fields_to_return
 
     @property
